@@ -6,10 +6,9 @@ import urllib.parse
 import hashlib
 import sys, os
 import xml.etree.ElementTree as ET
-from xml.dom import minidom
 import json
 import datetime
-import re
+import ssl
 
 __all__ = ["BBBApiClient", "ApiError"]
 
@@ -23,12 +22,13 @@ class ApiError(RuntimeError):
 
 
 class BBBApiClient:
-    def __init__(self, api, secret):
+    def __init__(self, api, secret, ssl_context=None):
         self.api = api.rstrip("/")
         self.secret = secret
+        self.ssl = ssl_context or ssl.create_default_context()
 
     def makeurl(self, command, **query):
-        query = urllib.parse.urlencode(query)
+        query = urllib.parse.urlencode({k:v for k,v in query.items() if v is not None})
         checksum = hashlib.sha1(
             (command + query + self.secret).encode("utf8")
         ).hexdigest()
@@ -39,7 +39,7 @@ class BBBApiClient:
 
     def call(self, command, **query):
         url = self.makeurl(command, **query)
-        with urllib.request.urlopen(url) as f:
+        with urllib.request.urlopen(url, context=self.ssl) as f:
             xml = f.read().decode("utf8")
         root = ET.fromstring(xml)
         if root.find("./returncode").text != "SUCCESS":
@@ -70,6 +70,9 @@ class BBBApiClient:
     def deleteRecordings(self, recordID):
         return self.call("deleteRecordings", recordID=recordID)
 
+    def sendChatMessage(self, meetingID, message, userName=None):
+        return self.call("sendChatMessage", meetingID=meetingID, message=message, userName=userName)
+
 
 def build_parser():
 
@@ -84,6 +87,10 @@ def build_parser():
         "--secret",
         help="API secretd (default: BBBCTL_SECRET or local config)",
     )
+    parser.add_argument(
+        "--insecure",
+        help="Skip TLS verification and accept self-signed or invalid SSL certificates as if they were valid",
+    )
 
     parser.add_argument(
         "--format",
@@ -125,7 +132,7 @@ def build_parser():
 
     cmd = meet_sub.add_parser("list", help="List meetings")
     cmd.add_argument("--sort", help="Sort by a specific key")
-    cmd.add_argument("--no-user", action="store_true", help="Do not show participatns")
+    cmd.add_argument("--no-user", action="store_true", help="Do not show participants")
     cmd.set_defaults(cmd=cmd_meet_list)
 
     cmd = meet_sub.add_parser("info", help="Show meeting details")
@@ -135,10 +142,10 @@ def build_parser():
     cmd = meet_sub.add_parser("create", help="Create meeting")
     cmd.add_argument("id", help="Meeting ID")
     cmd.add_argument("name", help="Meeting name")
-    cmd.add_argument("--record", help="Allow recodring?", action="store_true")
+    cmd.add_argument("--record", help="Allow recording?", action="store_true")
     cmd.add_argument(
         "--mod",
-        help="Directly create and print a join link for this username. Can be repeated.",
+        help="Directly create and print a moderator join link for this username. Can be repeated.",
         action="append",
     )
     cmd.set_defaults(cmd=cmd_meet_create)
@@ -160,8 +167,15 @@ def build_parser():
     cmd.add_argument("id", help="Meeting ID")
     cmd.set_defaults(cmd=cmd_meet_end)
 
+    cmd = meet_sub.add_parser("chat", help="Send a chat message into a running meeting (BBB 3.0)")
+    cmd.add_argument("id", help="Meeting ID. Can be 'BROADCAST' do broadcase a message to all running meetings")
+    cmd.add_argument("message", help="The message to send to chat")
+    cmd.add_argument("--name", help="The name that will be shown as the sender of the chat message", default="SYSTEM")
+    cmd.set_defaults(cmd=cmd_meet_chat)
+
     cmd = meet_sub.add_parser("nuke", help="End ALL meeting")
-    cmd.add_argument("--doit", help="Really end all meetings?", action="store_true")
+    cmd.add_argument("--doit", help="Disable dry-run mode and actually end meetings?", action="store_true")
+    cmd.add_argument("--ask", help="Ask for every meeting?", action="store_true")
     cmd.set_defaults(cmd=cmd_meet_nuke)
 
     return parser
@@ -176,6 +190,7 @@ config_paths = [
     "/etc/bigbluebutton/bbb-web.properties",
     "/usr/share/bbb-web/WEB-INF/classes/bigbluebutton.properties",
     "/var/lib/tomcat7/webapps/bigbluebutton/WEB-INF/classes/bigbluebutton.properties",
+    "./bbbctl.conf",
 ]
 
 
@@ -213,12 +228,17 @@ def main():
     )
 
     server = server.rstrip("/")
-    if '://' not in server:
+    if "://" not in server:
         server = "https://" + server
     if not server.endswith("/bigbluebutton/api"):
         server += "/bigbluebutton/api"
 
-    client = BBBApiClient(server, secret)
+    ctx = ssl.create_default_context()
+    if args.insecure:
+        ctx.check_hostname = False
+        ctx.verify_mode = ssl.CERT_NONE
+
+    client = BBBApiClient(server, secret, ssl_context=ctx)
 
     try:
         args.cmd(client, args)
@@ -249,7 +269,7 @@ def format_human(e, indent=""):
     result = indent + e.tag + ":"
     if e.tag.lower() in ("starttime", "endtime"):
         value = "%s (%s)" % (
-            datetime.datetime.utcfromtimestamp(float(value) / 1000),
+            datetime.datetime.fromtimestamp(float(value) / 1000, datetime.timezone.utc),
             value,
         )
     if value:
@@ -278,7 +298,19 @@ def format_json(e, *a, **ka):
 
 
 _should_be_list = ["attendees"]
-_should_be_number = ["createTime", "duration", "startTime", "endTime", "participantCount", "listenerCount", "voiceParticipantCount", "videoCount", "maxUsers", "moderatorCount"] 
+_should_be_number = [
+    "createTime",
+    "duration",
+    "startTime",
+    "endTime",
+    "participantCount",
+    "listenerCount",
+    "voiceParticipantCount",
+    "videoCount",
+    "maxUsers",
+    "moderatorCount",
+]
+
 
 def _unpack_xml(e):
     if e.tag in _should_be_list:
@@ -290,8 +322,8 @@ def _unpack_xml(e):
         return int(value)
     if not value:
         return None
-    if value in ('true', 'false'):
-        return value == 'true'
+    if value in ("true", "false"):
+        return value == "true"
     return value
 
 
@@ -351,7 +383,8 @@ def cmd_meet_create(api, args):
             link = api.getJoinLink(
                 meetingID=args.id,
                 fullName=name,
-                password=created.find("moderatorPW").text,
+                createTime=created.find("createTime").text,
+                role="MODERATOR",
             )
         print(name + ":", link)
 
@@ -365,6 +398,7 @@ def cmd_meet_join(api, args):
 
     if args.open:
         import webbrowser
+
         webbrowser.open_new_tab(link)
     else:
         print(link)
@@ -375,14 +409,19 @@ def cmd_meet_end(api, args):
     api.end(meetingID=args.id, password=pwd)
 
 
+def cmd_meet_chat(api, args):
+    if args.id == "BROADCAST":
+        meetings = [m.find("meetingID").text for m in api.getMeetings()]
+    else:
+        meetings = [args.id]
+
+    for meeting in meetings:
+        api.sendChatMessage(meeting, args.message, userName=args.name)
+
+
 def cmd_meet_nuke(api, args):
     meetings = list(api.getMeetings())
     for meeting in meetings:
-        if args.doit:
-            api.end(
-                meetingID=meeting.find("meetingID").text,
-                password=meeting.find("moderatorPW").text,
-            )
         print(
             "{dryrun}id={id} user={users} name={name!r}".format(
                 dryrun="" if args.doit else "(dry run) ",
@@ -391,6 +430,16 @@ def cmd_meet_nuke(api, args):
                 name=meeting.find("meetingName").text,
             )
         )
+        if args.doit:
+            if args.ask:
+                answer = input("End this meeting? [Yna] ").strip().lower()
+                if answer == "a":
+                    args.ask = False
+                elif answer in ("", "y"):
+                    pass
+                else:
+                    continue
+            api.end(meetingID=meeting.find("meetingID").text)
 
 
 if __name__ == "__main__":
